import logging
import os
import sys
import time
import traceback
from dataclasses import dataclass
from queue import Queue
from typing import Dict, Tuple, Any

import numpy as np

from era_5g_interface.channels import CallbackInfoServer, ChannelType, DATA_NAMESPACE, DATA_ERROR_EVENT
from era_5g_interface.dataclasses.control_command import ControlCommand, ControlCmdType
from era_5g_interface.interface_helpers import HeartbeatSender
from era_5g_interface.task_handler_internal_q import TaskHandlerInternalQ
from era_5g_server.server import NETAPP_STATUS_ADDRESS, NetworkApplicationServer, generate_application_heartbeat_data
from fcw_core.yolo_detector import YOLODetector
from fcw_service.collision_worker import CollisionWorker

logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
logger = logging.getLogger("FCW interface")

# Port of the 5G-ERA Network Application's server.
NETAPP_PORT = int(os.getenv("NETAPP_PORT", 5896))
# Input queue size.
NETAPP_INPUT_QUEUE = int(os.getenv("NETAPP_INPUT_QUEUE", 1))
# Event name for image error.
IMAGE_ERROR_EVENT = str("image_error")

EXTENDED_MEASURING = bool(os.getenv("EXTENDED_MEASURING", False))


@dataclass
class TaskAndWorker:
    """Class for task and worker."""

    task: TaskHandlerInternalQ
    worker: CollisionWorker


class Server(NetworkApplicationServer):
    """FCW server receives images from clients, manages tasks and workers, sends results to clients."""

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Constructor.

        Args:
            *args: NetworkApplicationServer arguments.
            **kwargs: NetworkApplicationServer arguments.
        """

        super().__init__(
            callbacks_info={
                "image_h264": CallbackInfoServer(ChannelType.H264, self.image_callback),
                "image_hevc": CallbackInfoServer(ChannelType.HEVC, self.image_callback),
                "image_jpeg": CallbackInfoServer(ChannelType.JPEG, self.image_callback),
            },
            **kwargs,
        )

        # List of registered tasks.
        self.tasks: Dict[str, TaskAndWorker] = dict()

        # Create Heartbeat sender
        self.heartbeat_sender = HeartbeatSender(NETAPP_STATUS_ADDRESS, self.generate_heartbeat_data)

    def generate_heartbeat_data(self):
        """Application heartbeat data generation using queue info and latencies."""

        latencies = []
        queue_occupancy = 0
        queue_size = 0
        for task_and_worker in self.tasks.values():
            queue_occupancy += task_and_worker.task.data_queue_occupancy()
            queue_size += task_and_worker.task.data_queue_size()
            latencies.extend(task_and_worker.worker.latency_measurements.get_latencies())
        avg_latency = 0
        if len(latencies) > 0:
            avg_latency = float(np.mean(np.array(latencies)))

        return generate_application_heartbeat_data(avg_latency, queue_size, queue_occupancy, len(self.tasks))

    def image_callback(self, sid: str, data: Dict[str, Any]) -> None:
        """Allows to receive decoded image using the websocket transport.

        Args:
            sid (str): Namespace sid.
            data (Dict[str, Any]): Data dict including decoded frame (data["frame"]) and send timestamp
                (data["timestamp"]).
        """

        eio_sid = self.get_eio_sid_of_data(sid)

        if eio_sid not in self.tasks:
            logger.error(f"Non-registered client {eio_sid} tried to send data")
            self.send_data({"message": "Non-registered client tried to send data"}, DATA_ERROR_EVENT, sid=sid)
            self.disconnect(sid)
            return

        if not self.tasks[eio_sid].worker.is_alive():
            logger.error(f"Worker is not alive, eio_sid {eio_sid}, sid {sid}")
            self.send_data({"message": f"Worker is not alive, eio_sid {eio_sid}, sid {sid}"}, DATA_ERROR_EVENT, sid=sid)
            self.disconnect(sid)
            return

        task = self.tasks[eio_sid].task
        task.store_data({"timestamp": data["timestamp"], "recv_timestamp": time.perf_counter_ns()}, data["frame"])

    def command_callback(self, command: ControlCommand, sid: str) -> Tuple[bool, str]:
        """Process initialization control command - create task, worker and start the worker.

        Args:
            command (ControlCommand): Control command to be processed.
            sid (str): Namespace sid.

        Returns:
            (initialized (bool), message (str)): If False, initialization failed.
        """

        eio_sid = self.get_eio_sid_of_control(sid)

        logger.info(f"Control command {command} processing: session id: {sid}")

        if command and command.cmd_type == ControlCmdType.INIT:
            # Check that initialization has not been called before.
            if eio_sid in self.tasks:
                logger.error(f"Client attempted to call initialization multiple times")
                self.send_command_error("Initialization has already been called before", sid)
                return False, "Initialization has already been called before"

            args = command.data
            config = {}
            camera_config = {}
            fps = 30
            viz = True
            viz_zmq_port = 5558
            if args:
                config = args.get("config", config)
                camera_config = args.get("camera_config", camera_config)
                fps = args.get("fps", fps)
                viz = args.get("viz", viz)
                viz_zmq_port = args.get("viz_zmq_port", viz_zmq_port)
                logger.info(f"Config: {config}")
                logger.info(f"Camera config: {camera_config}")
                logger.info(f"ZeroMQ visualization: {viz}, port: {viz_zmq_port}")

            # Queue with received images.
            image_queue = Queue(NETAPP_INPUT_QUEUE)

            task = TaskHandlerInternalQ(image_queue)

            try:
                # Create worker.
                worker = CollisionWorker(
                    image_queue=image_queue,
                    send_function=lambda results: self.send_data(
                        data=results, event="results",
                        sid=self.get_sid_of_data(eio_sid)
                    ),
                    config=config,
                    camera_config=camera_config,
                    fps=fps,
                    send_error_function=lambda message: self.send_data(
                        data=message, event=DATA_ERROR_EVENT, sid=self.get_sid_of_data(eio_sid)
                    ),
                    viz=viz,
                    viz_zmq_port=viz_zmq_port,
                    name=f"Collision Worker {eio_sid}",
                    daemon=True,
                )
            except Exception as ex:
                logger.error(f"Failed to create CollisionWorker: {repr(ex)}")
                logger.error(traceback.format_exc())
                self.send_command_error(f"Failed to create CollisionWorker: {repr(ex)}", sid)
                return False, f"Failed to create CollisionWorker: {repr(ex)}"

            self.tasks[eio_sid] = TaskAndWorker(task, worker)
            self.tasks[eio_sid].worker.start()
            t0 = time.perf_counter_ns()
            while True:
                if self.tasks[eio_sid].worker.is_alive():
                    break
                if time.perf_counter_ns() > t0 + 5 * 1.0e9:
                    logger.error(f"Timed out to start worker, eio_sid {eio_sid}, sid {sid}")
                    return False, f"Timed out to start worker"

            logger.info(f"Task handler and worker created and started: {eio_sid}")

        logger.info(
            f"Control command applied, eio_sid {eio_sid}, sid {sid}, "
            f"results sid {self.get_sid_of_data(eio_sid)}, command {command}"
        )
        return True, (
            f"Control command applied, eio_sid {eio_sid}, sid {sid}, results sid"
            f" {self.get_sid_of_data(eio_sid)}, command {command}"
        )

    def disconnect(self, sid: str) -> None:
        """Disconnects the client from DATA_NAMESPACE by sid.

        Args:
            sid (str): Namespace sid.
        """

        self._sio.disconnect(sid, DATA_NAMESPACE)

    def disconnect_callback(self, sid: str) -> None:
        """Called with client disconnection - deletes task and worker.

        Args:
            sid (str): Namespace sid.
        """

        eio_sid = self.get_eio_sid_of_data(sid)
        task_and_worker = self.tasks.get(eio_sid)
        if task_and_worker:
            task_and_worker.worker.stop()
            task_and_worker.worker.join()
            del self.tasks[eio_sid]
            del task_and_worker
            logger.info(f"Task handler and worker deleted: {eio_sid}")

        logger.info(f"Client disconnected from {DATA_NAMESPACE} namespace, eio_sid {eio_sid}, sid {sid}")


def main():
    """Main function."""

    # parser = argparse.ArgumentParser(description='Forward Collision Warning Service')
    # args = parser.parse_args()

    logger.info(f"The size of the queue set to: {NETAPP_INPUT_QUEUE}")

    logger.info("Initializing default object detector for faster first startup")
    detector = YOLODetector.from_dict({})
    del detector

    server = Server(port=NETAPP_PORT, host="0.0.0.0", extended_measuring=EXTENDED_MEASURING)

    try:
        server.run_server()
    except KeyboardInterrupt:
        logger.info("Terminating ...")


if __name__ == "__main__":
    main()



import time
from dataclasses import asdict
from multiprocessing import Queue
from queue import Empty
from threading import Thread, Event
from typing import Callable, Any

import zmq
from zmq import ZMQError

from era_5g_interface.interface_helpers import LatencyMeasurements
from fcw_core.detection import *
from fcw_core.sort import Sort, KalmanBoxTracker
from fcw_core.yolo_detector import YOLODetector
from fcw_core_utils.collision import *

logger = logging.getLogger(__name__)


class CollisionWorker(Thread):
    """FCW worker. Reads data from passed queue, performs FCW processing and returns results using callback."""

    def __init__(
        self,
        image_queue: Queue,
        send_function: Callable[[Dict[str, Any]], None],
        config: Dict,
        camera_config: Dict,
        fps: float,
        send_error_function: Callable[[Dict[str, Any]], None] = None,
        viz: bool = False,
        viz_zmq_port: int = 5558,
        **kw,
    ) -> None:
        """Constructor.

        Args:
            image_queue (Queue): The queue with all to-be-processed images.
            send_function (Callable[[Dict], None]): Callback used to send results.
            config (Dict): FCW config.
            camera_config (Dict): Camera config.
            fps (float): Framerate.
            send_error_function (Callable[[Dict], None]): Callback used to send errors.
            viz (bool): Enable visualization?
            viz_zmq_port (int): Visualization ZeroMQ port.
            **kw: Thread arguments.
        """

        super().__init__(**kw)

        self._stop_event = Event()
        self.image_queue = image_queue
        self._send_function = send_function
        self._send_error_function = send_error_function
        self._frame_id = 0
        self.latency_measurements: LatencyMeasurements = LatencyMeasurements()
        self._viz = viz

        logger.info("Initializing object detector")
        self._detector = YOLODetector.from_dict(config.get("detector", {}))
        logger.info("Initializing image tracker")
        self._tracker = Sort.from_dict(config.get("tracker", {}))
        logger.info("Initializing forward collision guard")
        self._guard = ForwardCollisionGuard.from_dict(config.get("fcw", {}))
        self._guard.dt = 1 / fps
        logger.info("Initializing camera calibration")
        self._camera = Camera.from_dict(camera_config)
        self._config = dict(config=config, camera_config=camera_config)

        # Visualization stuff.
        if self._viz:
            try:
                self._context = zmq.Context()
                self._socket = self._context.socket(zmq.PUB)
                logger.info(f"Publishing visualization on ZeroMQ tcp://*:{viz_zmq_port}")
                self._socket.bind("tcp://*:%s" % viz_zmq_port)
            except ZMQError as e:
                logger.error(f"Visualization was disabled on this worker: {e}")
                self._viz = False

    def stop(self) -> None:
        """Set stop event to stop FCW worker."""

        self._stop_event.set()

    def __del__(self) -> None:
        logger.info("Delete object detector")
        del self._detector

    def run(self) -> None:
        """FCW worker loop. Periodically reads images from python internal queue process them."""

        logger.info(f"{self.name} thread is running.")

        while not self._stop_event.is_set():
            # Get image and metadata from input queue.
            metadata: Dict[str, Any]
            image: np.ndarray
            try:
                metadata, image = self.image_queue.get(block=True, timeout=1)
            except Empty:
                continue
            # Store timestamp before processing.
            metadata["timestamp_before_process"] = time.perf_counter_ns()
            self._frame_id += 1
            # logger.info(f"Worker received frame id: {self.frame_id} {metadata['timestamp']}")
            try:
                detections = self._process_image(image)
                # Store timestamp after processing.
                metadata["timestamp_after_process"] = time.perf_counter_ns()
                # Generate results.
                results = self._generate_results(detections, metadata)
                # Send results via the provided callback.
                self._send_function(results)

                self.latency_measurements.store_latency(time.perf_counter_ns() - metadata["recv_timestamp"])

                if self._viz:
                    # If visualisation is enabled, send image with results over ZeroMQ.
                    self._send_image_with_results(image, results)

            except Exception as ex:
                logger.error(f"Exception with image processing ({type(ex)}): {repr(ex)}")
                if self._send_error_function:
                    self._send_error_function({"message": f"Exception with image processing ({type(ex)}): {repr(ex)}"})
                raise ex

        logger.info(f"{self.name} thread is stopping.")

    def _process_image(self, image: np.ndarray) -> Dict[int, KalmanBoxTracker]:
        """Process image by FCW.

        Args:
            image (np.ndarray): Image to be processed.

        Returns:
            Dictionary of KalmanBoxTrackers.
        """

        # Detect object in image.
        detections = self._detector.detect(image)
        # Get bounding boxes as numpy array.
        detections = detections_to_numpy(detections)
        # Update state of image trackers.
        self._tracker.update(detections)
        # Represent trackers as dict, {tid: KalmanBoxTracker, ...}.
        tracked_objects: Dict[int, KalmanBoxTracker] = {
            t.id: t for t in self._tracker.trackers if t.hit_streak > self._tracker.min_hits and t.time_since_update < 1
        }
        # Get 3D locations of objects.
        ref_points = get_reference_points(tracked_objects, self._camera, is_rectified=True)
        # Update state of objects in world.
        self._guard.update(ref_points)

        return tracked_objects

    def _send_image_with_results(self, image: np.ndarray, results: Dict[str, Any]) -> None:
        """Send a numpy array with metadata.

        Args:
            image (np.ndarray): Image.
            results (Dict[str, Any]): FCW results.
        """

        md = dict(dtype=str(image.dtype), shape=image.shape, results=dict(results, config=self._config))
        self._socket.send_json(md, 0 | zmq.SNDMORE)
        self._socket.send(image, 0, copy=True, track=False)

    def _generate_results(
        self, tracked_objects: Dict[int, KalmanBoxTracker], metadata: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Generates the results.

        Args:
            tracked_objects (Dict[int, KalmanBoxTracker]): The results of the detection.
            metadata (Dict[str, Any]): 5G-ERA Network Application specific metadata related to processed image.

        Returns:
            Dictionary of results.
        """

        # Get list of current offenses.
        dangerous_objects = self._guard.dangerous_objects()
        dangerous_detections = dict()

        # Get object statuses.
        object_statuses = list(self._guard.label_objects(include_distant=False))

        if tracked_objects is not None:
            for tid, t in tracked_objects.items():
                x1, y1, x2, y2 = t.get_state()[0]
                det = dict()
                det["bbox"] = [x1, y1, x2, y2]
                det["dangerous_distance"] = 0
                det["age"] = t.age
                det["hit_streak"] = t.hit_streak
                det["class"] = t.label
                det["class_name"] = self._detector.model.names[t.label]

                if tid in dangerous_objects.keys():
                    dist = Point(dangerous_objects[tid].location).distance(self._guard.vehicle_zone)
                    det["dangerous_distance"] = dist
                dangerous_detections[tid] = det

            # Make object statuses serializable - convert from shapely types.
            for i in range(len(object_statuses)):
                object_statuses[i].location = object_statuses[i].location.coords[0]
                object_statuses[i].path = [pts for pts in object_statuses[i].path.coords]
            object_statuses = [asdict(object_status) for object_status in object_statuses]

            return {
                "timestamp": metadata.get("timestamp", 0),
                "recv_timestamp": metadata.get("recv_timestamp", 0),
                "timestamp_before_process": metadata["timestamp_before_process"],
                "timestamp_after_process": metadata["timestamp_after_process"],
                "send_timestamp": time.perf_counter_ns(),
                "dangerous_detections": dangerous_detections,
                "objects": object_statuses,
            }



import argparse
import logging
import sys
import threading
import time
from queue import Full, Queue
from typing import Optional, Dict, Any
import av
from av.container.output import OutputContainer
from av.stream import Stream
import cv2
import numpy
import zmq
from zmq import Socket, Context

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("FCW visualization")

from fcw_core.vizualization import *


def recv_array(socket: Socket, flags=0, copy=True, track=False) -> (Dict[str, Any], np.ndarray):
    """recv a numpy array"""
    try:
        md = socket.recv_json(flags=flags)
        msg = socket.recv(flags=flags, copy=copy, track=track)
        buf = memoryview(msg)
        image = numpy.frombuffer(buf, dtype=md["dtype"])
        return md["results"], image.reshape(md["shape"])
    except zmq.error.Again as e:
        logger.debug("Missing visualization data!")
        return {}, None


def mark_vehicles(
    size: tuple, objects: list, camera: Camera, marker: Image, anchor: tuple = (0, 0), to_rectified=False
):
    image = Image.new("RGBA", size, color=(0, 0, 0, 0))
    ax, ay = anchor
    # loc (N,2) xy
    for o in objects:
        X = np.atleast_2d([o["location"][0], o["location"][1], 0])
        scr_loc, _ = camera.project_points(X, near=1, to_rectified=to_rectified)
        if scr_loc.shape[0] > 0:
            x, y = scr_loc[0]
            image.paste(marker, (int(x - ax), int(y - ay)))
    return image


def draw_world_objects(
    size: tuple,
    camera: Camera,
    objects: list,
    to_rectified=False,
):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    for o in objects:
        X = np.atleast_2d([o["location"][0], o["location"][1], 0])
        scr_loc, _ = camera.project_points(X, to_rectified=to_rectified)
        if scr_loc.size > 0:
            x, y = scr_loc[0]
            draw.line([(x - 10, y), (x + 10, y)], fill=(255, 255, 0, 128), width=3)
            draw.line([(x, y - 10), (x, y + 10)], fill=(255, 255, 0, 128), width=3)

        X = np.array(o["path"])
        n = X.shape[0]
        X = np.hstack([X, np.zeros((n, 1))])
        scr_loc, _ = camera.project_points(X, near=5, to_rectified=to_rectified)
        scr_loc = list(map(tuple, scr_loc))
        draw.line(scr_loc, fill=(0, 255, 0, 255), width=1)

    return image


def draw_image_trackers(
    size: tuple,
    trackers: list,
):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    for t in trackers:
        x1, y1, x2, y2 = t["bbox"]
        color = (0, 255, 0, 64)
        outline = (0, 255, 0, 128)
        if t["age"] < 3 or t["hit_streak"] == 0:  # TODO: call it.is_reliable()
            color = (255, 255, 0, 32)
            outline = None
        draw.rectangle((x1, y1, x2, y2), fill=color, outline=outline)

    return image


recv_queue = Queue(5)
context: Context = zmq.Context()
socket: Socket = context.socket(zmq.SUB)
socket.setsockopt(zmq.SUBSCRIBE, b"")
socket.setsockopt(zmq.RCVTIMEO, 2000)


def recv_results_with_images():
    while True:
        results, image = recv_array(socket)
        try:
            recv_queue.put_nowait((results, image))
        except Full:
            logger.info(f"Results dropped {results}")


def main(args=None):
    parser = argparse.ArgumentParser(description="Visualization of Forward Collision Warning Service")
    parser.add_argument("-z", "--zmq_port", type=str, help="ZeroMQ port", default="5558")
    parser.add_argument(
        "-u", "--rtsp_port", type=str, help="RTSP port, address is rtsp://localhost:{rtsp_port}/video", default="8554"
    )
    args = parser.parse_args()

    socket.connect("tcp://localhost:%s" % args.zmq_port)

    camera: Optional[Camera] = None
    config: Optional[Dict] = None
    output = None

    recv_thread = threading.Thread(target=recv_results_with_images, daemon=True)
    recv_thread.start()

    while True:
        try:
            results, image = recv_queue.get()
            if results is None or image is None:
                cv2.destroyAllWindows()
                continue
            if not config or config != results["config"]:
                config = results["config"]
                if "camera_config" not in config:
                    config = None
                    continue
                logger.info("Initializing camera calibration")
                camera = Camera.from_dict(config["camera_config"])
                # logo = cog_logo((64, 64))
                coord_sys = draw_world_coordinate_system(camera.rectified_size, camera)
                coord_sys.putalpha(64)
                if type(config["config"]["fcw"].get("danger_zone")) == dict:
                    zone = Polygon(list(config["config"]["fcw"].get("danger_zone").values()))
                else:
                    zone = Polygon(config["config"]["fcw"].get("danger_zone"))
                danger_zone = draw_danger_zone(camera.rectified_size, camera, zone)
                horizon = draw_horizon(camera.rectified_size, camera, width=1, fill=(255, 255, 0, 64))
                marker, marker_anchor = vehicle_marker_image(scale=3)

                if output is not None:
                    output.close()
                # TODO: Check RTSP server is running
                output: OutputContainer = av.open(
                    f"rtsp://localhost:{args.rtsp_port}/video",
                    mode="w",
                    format="rtsp",
                    options={"rtsp_transport": "tcp"},
                    timeout=2,
                )
                output.flags |= output.flags.NONBLOCK
                out_stream: Stream = output.add_stream("h264", 30)
                logger.debug(out_stream.codec_context.is_open)
                out_stream.pix_fmt = "yuv420p"
                out_stream.options = {"preset": "ultrafast", "tune": "zerolatency", "crf": "20"}
                out_stream.width = image.shape[1]
                out_stream.height = image.shape[0]

            logger.debug(results["dangerous_detections"])
            logger.debug(results["objects"])
            logger.debug("--------")

            base_undistorted = Image.fromarray(image[..., ::-1], "RGB").convert("RGBA")
            # Layers showing various information
            sz = base_undistorted.size
            layers = [
                (coord_sys, None),
                (danger_zone, None),
                (horizon, None),
                (draw_image_trackers(sz, list(results["dangerous_detections"].values())), None),
                (draw_world_objects(sz, camera, list(results["objects"]), to_rectified=True), None),
            ]
            # Compose layers together
            compose_layers(base_undistorted, *layers)
            object_statuses: List[ObjectStatus] = []
            for object_status_str in results["objects"]:
                object_status = ObjectStatus(
                    id=object_status_str["id"],
                    distance=object_status_str["distance"],
                    location=object_status_str["location"],
                    path=object_status_str["path"],
                    is_in_danger_zone=object_status_str["is_in_danger_zone"],
                    crosses_danger_zone=object_status_str["crosses_danger_zone"],
                    time_to_collision=object_status_str["time_to_collision"],
                )
                object_statuses.append(object_status)
            w1, h1 = base_undistorted.size
            compose_layers(
                base_undistorted,  # Original image
                (tracking_info((w1, 16), object_statuses), (0, 0)),
                (
                    mark_vehicles(
                        camera.image_size, list(results["objects"]), camera, marker, marker_anchor, to_rectified=True
                    ),
                    (0, 0),
                ),
                # (logo, (8, 16 + 8)),
            )
            # Convert to OpenCV for display
            cv_image = np.array(base_undistorted.convert("RGB"))[..., ::-1]

            try:
                out_frame = av.VideoFrame.from_ndarray(cv_image, format="bgr24")
                out_packet = out_stream.encode(out_frame)
                logger.info(out_packet)
                output.mux(out_packet)
                # time.sleep(1 / float(in_stream.framerate))
            except Exception as ex:
                config = None
                logger.error(repr(ex))
            # try:
            # Display the image
            #    cv2.imshow("FCW", cv_image)
            #    cv2.waitKey(1)
            # except Exception as ex:
            #    logger.debug(repr(ex))
        except KeyboardInterrupt:
            logger.info("Terminating ...")
            break


if __name__ == "__main__":
    main()
